/*
 *  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * popup.c - functions to deal with popup window.
 */

#include "vncviewer.h"

#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Toggle.h>

Widget popup, fullScreenToggle;

static void popupFixer(Widget wid) {
	Window rr, cr; 
	unsigned int m;
  	int x0 = 500, y0 = 500;
	int xr, yr, wxr, wyr;
	Dimension ph;
	if (XQueryPointer(dpy, DefaultRootWindow(dpy), &rr, &cr, &xr, &yr, &wxr, &wyr, &m)) {
		x0 = xr;
		y0 = yr;
	}
  	XtPopup(wid, XtGrabNone);
	XtVaGetValues(wid, XtNheight, &ph, NULL);
	if (y0 + (int) ph > dpyHeight) {
		y0 = dpyHeight - (int) ph;
		if (y0 < 0) {
			y0 = 0;
		}
	}
	XtMoveWidget(wid, x0, y0);
}

void
ShowPopup(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
  if (appData.popupFix) {
	popupFixer(popup);
  } else {
	XtMoveWidget(popup, event->xbutton.x_root, event->xbutton.y_root);
  	XtPopup(popup, XtGrabNone);
  }
  if (appData.grabAll) {
  	XRaiseWindow(dpy, XtWindow(popup));
  }
  XSetWMProtocols(dpy, XtWindow(popup), &wmDeleteWindow, 1);
}

void
HidePopup(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
  XtPopdown(popup);
}


static XtResource resources[] = {
  {
    "type", "Type", XtRString, sizeof(String), 0, XtRString,
    (XtPointer) "command",
  },
};

void
CreatePopup()
{
  Widget buttonForm, button, prevButton = NULL;
  int i;
  char buttonName[12];
  String buttonType;

  popup = XtVaCreatePopupShell("popup", transientShellWidgetClass, toplevel,
			       NULL);

  buttonForm = XtVaCreateManagedWidget("buttonForm", formWidgetClass, popup,
				       NULL);

  if (appData.popupButtonCount > 100) {
    fprintf(stderr,"Too many popup buttons\n");
    exit(1);
  }

  for (i = 1; i <= appData.popupButtonCount; i++) {
    sprintf(buttonName, "button%d", i);
    XtVaGetSubresources(buttonForm, (XtPointer)&buttonType, buttonName,
			"Button", resources, 1, NULL);

    if (strcmp(buttonType, "command") == 0) {
      button = XtVaCreateManagedWidget(buttonName, commandWidgetClass,
				       buttonForm, NULL);
      XtVaSetValues(button, XtNfromVert, prevButton,
		    XtNleft, XawChainLeft, XtNright, XawChainRight, NULL);
    } else if (strcmp(buttonType, "toggle") == 0) {
      button = XtVaCreateManagedWidget(buttonName, toggleWidgetClass,
				       buttonForm, NULL);
      XtVaSetValues(button, XtNfromVert, prevButton,
		    XtNleft, XawChainLeft, XtNright, XawChainRight, NULL);
    } else {
      fprintf(stderr,"unknown button type '%s'\n",buttonType);
    }
    prevButton = button;
  }
}


Widget scaleN;

void
ShowScaleN(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
  if (appData.popupFix) {
	popupFixer(scaleN);

  } else {
	XtMoveWidget(scaleN, event->xbutton.x_root, event->xbutton.y_root);
  	XtPopup(scaleN, XtGrabNone);
  }
  if (appData.grabAll) {
  	XRaiseWindow(dpy, XtWindow(scaleN));
  }
  XSetWMProtocols(dpy, XtWindow(scaleN), &wmDeleteWindow, 1);
}

void
HideScaleN(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
  XtPopdown(scaleN);
}


void
CreateScaleN()
{
  Widget buttonForm, button, prevButton = NULL;
  int i;
  char buttonName[32];
  String buttonType;

  scaleN = XtVaCreatePopupShell("scaleN", transientShellWidgetClass, toplevel,
			       NULL);

  buttonForm = XtVaCreateManagedWidget("buttonForm", formWidgetClass, scaleN,
				       NULL);

  for (i = 0; i <= 6; i++) {
    sprintf(buttonName, "button%d", i);
    XtVaGetSubresources(buttonForm, (XtPointer)&buttonType, buttonName,
			"Button", resources, 1, NULL);

    button = XtVaCreateManagedWidget(buttonName, commandWidgetClass,
				       buttonForm, NULL);
    XtVaSetValues(button, XtNfromVert, prevButton,
		    XtNleft, XawChainLeft, XtNright, XawChainRight, NULL);
    prevButton = button;
  }
}

int filexfer_sock = -1;
int filexfer_listen = -1;

void HideFile(Widget w, XEvent *event, String *params, Cardinal *num_params) {
	if (filexfer_sock >= 0) {
		close(filexfer_sock);
		filexfer_sock = -1;
	}
	if (filexfer_listen >= 0) {
		close(filexfer_listen);
		filexfer_listen = -1;
	}
}

extern int use_loopback;
time_t start_listen = 0;

void ShowFile(Widget w, XEvent *event, String *params, Cardinal *num_params) {
	int i, port0 = 7200, port, sock = -1;
	char *cmd, *jar;
	char fmt[] = "java -cp '%s' VncViewer HOST localhost PORT %d delayAuthPanel yes ignoreMSLogonCheck yes disableSSL yes ftpOnly yes graftFtp yes &";

	if (getenv("SSVNC_ULTRA_FTP_JAR")) {
		jar = getenv("SSVNC_ULTRA_FTP_JAR");
		cmd = (char *) malloc(strlen(fmt) + strlen(jar) + 10);
	} else {
		fprintf(stderr, "Cannot find UltraVNC FTP jar file.\n");
		return;
	}
	use_loopback = 1;
	for (i = 0; i < 100; i++) {
		port = port0 + i;
		sock = ListenAtTcpPort(port);
		if (sock >= 0) {
			fprintf(stderr, "listening for filexfer on port: %d sock: %d\n", port, sock);
			break;
		}
	}
	use_loopback = 0;
	if (sock >= 0) {
		int st;
		pid_t pid = fork();
		if (pid < 0) {
			free(cmd);
			return;
		} else if (pid == 0) {
			int i;
			sprintf(cmd, fmt, jar, port);
			for (i = 3; i < 100; i++) {
				close(i);
			}
			fprintf(stderr, "\n-- Experimental UltraVNC File Transfer --\n\nRunning cmd:\n\n  %s\n\n", cmd);
			system(cmd);
			exit(0);
		}
		waitpid(pid, &st, 0);
		start_listen = time(NULL);
	}
	free(cmd);
	filexfer_listen = sock;
}

Widget chat, entry, text;

static int chat_visible = 0;

void
ShowChat(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	if (appData.termChat) {
		return;
	}
	if (! chat_visible) {
		XtPopup(chat, XtGrabNone);
		chat_visible = 1;
		XSetWMProtocols(dpy, XtWindow(chat), &wmDeleteWindow, 1);
		XSync(dpy, False);
		usleep(200 * 1000);
	}
}

void hidechat(void) {
	if (appData.termChat) {
		return;
	}
	if (chat_visible) {
		XtPopdown(chat);
		chat_visible = 0;
		XSync(dpy, False);
		usleep(200 * 1000);
	}
}

void HideChat(Widget w, XEvent *event, String *params, Cardinal *num_params) {
	SendTextChatClose();
	hidechat();
}

void dismiss_proc(Widget w, XtPointer client_data, XtPointer call_data) {
	SendTextChatClose();
	hidechat();
}

extern void printChat(char *, Bool);

void CheckTextInput(void) {
	Arg args[2];
	String str;
	int len;

	if (appData.termChat) {
		return;
	}

	XtSetArg(args[0], XtNstring, &str);
	XtGetValues(entry, args, 1);
	if (str == NULL || str[0] == '\0') {
		return;
	} else {
		len = strlen(str);
		if (len <= 0) {
			return;
		}
		if (str[len-1] == '\n') {
			char *s = strdup(str);
			if (s) {
				SendTextChat(s);
				printChat("Send: ", True);
				printChat(s, True);
				XtVaSetValues(entry, XtNtype, XawAsciiString, XtNstring, "", NULL);
				free(s);
			}
		}
	}
}

void AppendChatInput0(char *in) {
	Arg args[10];
	int n;
	String str;
	int len;
	static char *s = NULL;
	static int slen = -1;
	XawTextPosition pos;

	fprintf(stderr, "AppendChatInput: in= '%s'\n", in);

	XtSetArg(args[0], XtNstring, &str);
	XtGetValues(text, args, 1);
	fprintf(stderr, "AppendChatInput: str='%s'\n", str);
	
	len = strlen(str) + strlen(in);

	if (slen <= len) {
		slen = 2 * (len + 10);
		if (s) free(s);
		s = (char *) malloc(slen+1);
	}
	
	s[0] = '\0';
	strcat(s, str);
	strcat(s, in);
	fprintf(stderr, "AppendChatInput s=  '%s'\n", s);
	pos = (XawTextPosition) (len-1);
	n = 0;
	XtSetArg(args[n], XtNtype, XawAsciiString);	n++;
	XtSetArg(args[n], XtNstring, s);		n++;
	XtSetArg(args[n], XtNdisplayPosition, pos);	n++;
	XtSetArg(args[n], XtNinsertPosition, pos);	n++;
	XtSetValues(text, args, n);
	fprintf(stderr, "AppendChatInput done\n");
}

void AppendChatInput(char *in) {
	int len;
	XawTextPosition beg, end;
	static XawTextPosition pos = 0;
	XawTextBlock txt;

	if (appData.termChat) {
		return;
	}

	XawTextSetInsertionPoint(text, pos);
	beg = XawTextGetInsertionPoint(text);
	end = beg;
	//fprintf(stderr, "AppendChatInput: pos=%d in= '%s'\n", beg, in);

	txt.firstPos = 0;
	txt.length = strlen(in);
	txt.ptr = in;
	txt.format = FMT8BIT;

	XawTextReplace(text, beg, end, &txt);
	XawTextSetInsertionPoint(text, beg + txt.length);

	pos = XawTextGetInsertionPoint(text);
	//fprintf(stderr, "AppendChatInput done pos=%d\n", pos);
}

static char errorbuf[1] = {0};

void CreateChat(void) {

	Widget myform, dismiss;
	int i, n;
	Dimension w = 400, h = 300;
	Dimension x = 33, y = 33;
	Arg args[10];

	chat = XtVaCreatePopupShell("chat", topLevelShellWidgetClass, toplevel, XtNmappedWhenManaged, False, NULL);

	myform = XtVaCreateManagedWidget("myform", formWidgetClass, chat, NULL);

	text = XtVaCreateManagedWidget("text", asciiTextWidgetClass, myform,
	    XtNresize, XawtextResizeBoth, XtNresizable, True, XtNwrap, XawtextWrapWord,
	    XtNscrollHorizontal, XawtextScrollNever, XtNscrollVertical, XawtextScrollAlways,
	    XtNwidth, w, XtNheight, h, XtNdisplayCaret, False,
	    XtNeditType, XawtextAppend, XtNtype, XawAsciiString,
	    XtNuseStringInPlace, False, NULL);

	entry = XtVaCreateManagedWidget("entry", asciiTextWidgetClass, myform,
	    XtNresize, XawtextResizeWidth, XtNresizable, True, XtNwrap, XawtextWrapWord,
	    XtNscrollHorizontal, XawtextScrollNever, XtNscrollVertical, XawtextScrollNever,
	    XtNheight, 20, XtNwidth, 400, XtNfromVert, text,
	    XtNdisplayCaret, True, XtNeditType, XawtextEdit, NULL);

	dismiss = XtVaCreateManagedWidget("dismiss", commandWidgetClass, myform, XtNlabel, "Close Chat", XtNfromVert, entry, NULL);

	AppendChatInput("");

	XtAddCallback(dismiss, XtNcallback, dismiss_proc, NULL);

	XtRealizeWidget(chat);
}
